@@ -1,5 +1,5 @@
 /*
- *  $Id: netif.c 208 2008-07-27 16:15:45Z sten $
+ *  $Id: netif.c 318 2008-09-21 23:22:28Z sten $
  */
 
 #include "main.h"
@@ -54,8 +54,16 @@
 #endif /* HAVE_NET_IF_TRUNK_H */
 
 
-#if HAVE_LINUX_IF_BRIDGE_H
+#ifdef HAVE_LINUX_IF_BONDING_H
+#include <linux/if_bonding.h>
+#endif /* HAVE_LINUX_IF_BONDING_H */
+
+#ifdef HAVE_LINUX_IF_BRIDGE_H
 #include <linux/if_bridge.h>
+#ifndef SYSFS_BRIDGE_PORT_SUBDIR
+#define SYSFS_BRIDGE_PORT_SUBDIR "brif"
+#endif
+#define BRIDGE_MAX_PORTS 1024
 #endif /* HAVE_LINUX_IF_BRIDGE_H */
 
 #if HAVE_NET_IF_BRIDGEVAR_H
@@ -78,10 +86,15 @@
 #include <net80211/ieee80211_ioctl.h>
 #endif /* HAVE_NET80211_IEEE80211_IOCTL_H */
 
-#define SYSFS_VIRTUAL		"/sys/devices/virtual/net"
+#ifdef HAVE_SYSFS
+#define SYSFS_CLASS_NET		"/sys/class/net"
 #define SYSFS_PATH_MAX		256
+#endif
+
+#ifdef HAVE_PROC_SYS_NET
 #define PROCFS_FORWARD_IPV4	"/proc/sys/net/ipv4/conf/all/forwarding"
 #define PROCFS_FORWARD_IPV6	"/proc/sys/net/ipv6/conf/all/forwarding"
+#endif
 
 #ifdef AF_PACKET
 #define NETIF_AF    AF_PACKET
@@ -91,14 +104,14 @@
 
 int netif_wireless(int sockfd, struct ifaddrs *ifaddr, struct ifreq *);
 int netif_type(int sockfd, struct ifaddrs *ifaddr, struct ifreq *);
-void netif_bond(int sockfd, struct netif *, struct netif *);
-void netif_bridge(int sockfd, struct netif *, struct netif *);
-int netif_addrs(struct ifaddrs *, struct netif *);
-int netif_forwarding();
+void netif_bond(int sockfd, struct netif *, struct netif *, struct ifreq *);
+void netif_bridge(int sockfd, struct netif *, struct netif *, struct ifreq *);
+int netif_addrs(struct ifaddrs *, struct netif *, struct sysinfo *sysinfo);
+void netif_forwarding(struct sysinfo *);
 
 
 // create netifs for a list of interfaces
-uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
+uint16_t netif_fetch(int ifc, char *ifl[], struct sysinfo *sysinfo,
 		    struct netif **mnetifs) {
 
     int sockfd, af = AF_INET;
@@ -120,12 +133,16 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
     sockfd = my_socket(af, SOCK_DGRAM, 0);
 
     if (getifaddrs(&ifaddrs) < 0) {
-	my_log(0, "address detection failed: %s", strerror(errno));
-	close(sockfd);
+	my_log(CRIT, "address detection failed: %s", strerror(errno));
+	(void) close(sockfd);
 	return(0);
     }
 
     for (ifaddr = ifaddrs; ifaddr != NULL; ifaddr = ifaddr->ifa_next) {
+	// skip interfaces without addresses
+	if (ifaddr->ifa_addr == NULL)
+	    continue;
+
 	// only handle datalink addresses
 	if (ifaddr->ifa_addr->sa_family == NETIF_AF)
 	    count++;
@@ -134,7 +151,7 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
     // allocate memory
     netifs = realloc(*mnetifs, sizeof(struct netif) * count);
     if (netifs == NULL) {
-	my_log(3, "unable to allocate netifs");
+	my_log(INFO, "unable to allocate netifs");
 	goto cleanup;
     }
     *mnetifs = netifs;
@@ -143,8 +160,18 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
     memset(netifs, 0, sizeof(struct netif) * count);
     count = 0;
 
+    // default to CAP_HOST
+    sysinfo->cap = CAP_HOST;
+    sysinfo->cap_active = CAP_HOST;
+
     for (ifaddr = ifaddrs; ifaddr != NULL; ifaddr = ifaddr->ifa_next) {
 
+	// skip interfaces without addresses
+	if (ifaddr->ifa_addr == NULL) {
+	    my_log(INFO, "skipping interface %s", ifaddr->ifa_name);
+	    continue;
+	}
+
 	// only handle datalink addresses
 	if (ifaddr->ifa_addr->sa_family != NETIF_AF)
 	    continue;
@@ -155,21 +182,21 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
 
 	// prepare ifr struct
 	memset(&ifr, 0, sizeof(ifr));
-	strncpy(ifr.ifr_name, ifaddr->ifa_name, IFNAMSIZ);
+	strlcpy(ifr.ifr_name, ifaddr->ifa_name, sizeof(ifr.ifr_name));
 
 
 	// skip non-ethernet interfaces
 #ifdef AF_PACKET
 	memcpy(&saddrll, ifaddr->ifa_addr, sizeof(saddrll));
 	if (saddrll.sll_hatype != ARPHRD_ETHER) {
-	    my_log(3, "skipping interface %s", ifaddr->ifa_name);
+	    my_log(INFO, "skipping interface %s", ifaddr->ifa_name);
 	    continue;
 	}
 #endif
 #ifdef AF_LINK
 	memcpy(&saddrdl, ifaddr->ifa_addr, sizeof(saddrdl));
 	if (saddrdl.sdl_type != IFT_ETHER) {
-	    my_log(3, "skipping interface %s", ifaddr->ifa_name);
+	    my_log(INFO, "skipping interface %s", ifaddr->ifa_name);
 	    continue;
 	}
 #endif
@@ -180,7 +207,7 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
 
 	// skip wireless interfaces
 	if (netif_wireless(sockfd, ifaddr, &ifr) == 0) {
-	    my_log(3, "skipping wireless interface %s", ifaddr->ifa_name);
+	    my_log(INFO, "skipping wireless interface %s", ifaddr->ifa_name);
 	    sysinfo->cap |= CAP_WLAN; 
 	    sysinfo->cap_active |= (enabled == 1) ? CAP_WLAN : 0;
 	    continue;
@@ -190,27 +217,27 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
 	type = netif_type(sockfd, ifaddr, &ifr);
 
 	if (type == NETIF_REGULAR) { 
-	    my_log(2, "found ethernet interface %s", ifaddr->ifa_name);
+	    my_log(INFO, "found ethernet interface %s", ifaddr->ifa_name);
 	} else if (type == NETIF_BONDING) {
-	    my_log(2, "found bond interface %s", ifaddr->ifa_name);
+	    my_log(INFO, "found bond interface %s", ifaddr->ifa_name);
 	} else if (type == NETIF_BRIDGE) {
-	    my_log(2, "found bridge interface %s", ifaddr->ifa_name);
+	    my_log(INFO, "found bridge interface %s", ifaddr->ifa_name);
 	    sysinfo->cap |= CAP_BRIDGE; 
 	    sysinfo->cap_active |= (enabled == 1) ? CAP_BRIDGE : 0;
 	} else if (type == NETIF_INVALID) {
-	    my_log(3, "skipping interface %s", ifaddr->ifa_name);
+	    my_log(INFO, "skipping interface %s", ifaddr->ifa_name);
 	    continue;
 	}
 
 
 	// skip interfaces that are down
 	if (enabled == 0) {
-	    my_log(3, "skipping interface %s (down)", ifaddr->ifa_name);
+	    my_log(INFO, "skipping interface %s (down)", ifaddr->ifa_name);
 	    continue;
 	}
 
 
-	my_log(3, "adding interface %s", ifaddr->ifa_name);
+	my_log(INFO, "adding interface %s", ifaddr->ifa_name);
 
 	// create netif
 	if (netif == NULL)
@@ -225,9 +252,14 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
 #ifdef AF_LINK
 	netif->index = saddrdl.sdl_index;
 #endif
-	strncpy(netif->name, ifaddr->ifa_name, IFNAMSIZ);
+	strlcpy(netif->name, ifaddr->ifa_name, sizeof(netif->name));
 	netif->type = type;
 
+#ifdef SIOCGIFDESCR
+	ifr.ifr_data = (caddr_t)&netif->description;
+	ioctl(sockfd, SIOCGIFDESCR, &ifr);
+#endif
+
 	// update linked list
 	if (netif_prev != NULL)
 	    netif_prev->next = netif;
@@ -242,12 +274,12 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
 
 	switch(netif->type) {
 	    case NETIF_BONDING:
-		my_log(3, "detecting %s subifs", netif->name);
-		netif_bond(sockfd, netifs, netif);
+		my_log(INFO, "detecting %s settings", netif->name);
+		netif_bond(sockfd, netifs, netif, &ifr);
 		break;
 	    case NETIF_BRIDGE:
-		my_log(3, "detecting %s subifs", netif->name);
-		netif_bridge(sockfd, netifs, netif);
+		my_log(INFO, "detecting %s settings", netif->name);
+		netif_bridge(sockfd, netifs, netif, &ifr);
 		break;
 	    default:
 		break;
@@ -255,18 +287,15 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
     }
 
     // add addresses to netifs
-    my_log(3, "fetching addresses for all interfaces");
-    if (netif_addrs(ifaddrs, netifs) == EXIT_FAILURE) {
-	my_log(0, "unable fetch interface addresses");
+    my_log(INFO, "fetching addresses for all interfaces");
+    if (netif_addrs(ifaddrs, netifs, sysinfo) == EXIT_FAILURE) {
+	my_log(CRIT, "unable fetch interface addresses");
 	count = 0;
 	goto cleanup;
     }
 
     // check for forwarding
-    if (netif_forwarding() == 1) {
-	sysinfo->cap |= CAP_ROUTER; 
-	sysinfo->cap_active |= CAP_ROUTER; 
-    };
+    netif_forwarding(sysinfo);
 
     // use the first mac as chassis id
     memcpy(&sysinfo->hwaddr, &netifs->hwaddr, ETHER_ADDR_LEN);
@@ -278,7 +307,7 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
 	for (j=0; j < ifc; j++) {
 	    netif = netif_byname(netifs, ifl[j]);
 	    if (netif == NULL) {
-		my_log(0, "interface %s is invalid", ifl[j]);
+		my_log(CRIT, "interface %s is invalid", ifl[j]);
 	    } else {
 		netif->argv = 1;
 		count++;
@@ -288,13 +317,13 @@ uint16_t netif_list(int ifc, char *ifl[], struct sysinfo *sysinfo,
 	    count = 0;
 
     } else if (count == 0) {
-	my_log(0, "no valid interface found");
+	my_log(CRIT, "no valid interface found");
     }
 
     // cleanup
 cleanup:
     freeifaddrs(ifaddrs);
-    close(sockfd);
+    (void) close(sockfd);
 
     return(count);
 };
@@ -307,7 +336,7 @@ int netif_wireless(int sockfd, struct ifaddrs *ifaddr, struct ifreq *ifr) {
     struct iwreq iwreq;
 
     memset(&iwreq, 0, sizeof(iwreq));
-    strncpy(iwreq.ifr_name, ifaddr->ifa_name, IFNAMSIZ);
+    strlcpy(iwreq.ifr_name, ifaddr->ifa_name, sizeof(iwreq.ifr_name));
 
     return(ioctl(sockfd, SIOCGIWNAME, &iwreq));
 #endif
@@ -318,7 +347,7 @@ int netif_wireless(int sockfd, struct ifaddrs *ifaddr, struct ifreq *ifr) {
     u_int8_t i_data[32];
 
     memset(&ireq, 0, sizeof(ireq));
-    strncpy(ireq.i_name, ifaddr->ifa_name, sizeof(ireq.i_name));
+    strlcpy(ireq.i_name, ifaddr->ifa_name, sizeof(ireq.i_name));
     ireq.i_data = &i_data;
 
     ireq.i_type = IEEE80211_IOC_SSID;
@@ -342,19 +371,27 @@ int netif_wireless(int sockfd, struct ifaddrs *ifaddr, struct ifreq *ifr) {
 // detect interface type
 int netif_type(int sockfd, struct ifaddrs *ifaddr, struct ifreq *ifr) {
 
-#if HAVE_LINUX_ETHTOOL_H
+#ifdef HAVE_SYSFS
     char path[SYSFS_PATH_MAX];
     struct stat sb;
+#endif
 
+#if HAVE_LINUX_ETHTOOL_H
     struct ethtool_drvinfo drvinfo;
-
     memset(&drvinfo, 0, sizeof(drvinfo));
-    sprintf(path, "%s/%s", SYSFS_VIRTUAL, ifaddr->ifa_name); 
+#endif
 
-    // accept physical devices
-    if (stat(path, &sb) != 0)
-	return(NETIF_REGULAR);
+#ifdef HAVE_SYSFS
+    if (snprintf(path, SYSFS_PATH_MAX,
+	    SYSFS_CLASS_NET "/%s/device", ifaddr->ifa_name) > 0) {
 
+	// accept physical devices via sysfs
+	if (stat(path, &sb) == 0)
+	    return(NETIF_REGULAR);
+    }
+#endif
+
+#if HAVE_LINUX_ETHTOOL_H
     // use ethtool to detect various drivers
     drvinfo.cmd = ETHTOOL_GDRVINFO;
     ifr->ifr_data = (caddr_t)&drvinfo;
@@ -370,6 +407,11 @@ int netif_type(int sockfd, struct ifaddrs *ifaddr, struct ifreq *ifr) {
 	} else if (strcmp(drvinfo.driver, "tun") == 0) {
 	    return(NETIF_REGULAR);
 	}
+
+	// if we don't have sysfs, we'll accept interfaces
+	// which support ethtool (aka wing it)
+	if (stat(SYSFS_CLASS_NET, &sb) == -1)
+	    return(NETIF_REGULAR);
     }
 
     // we don't want the rest
@@ -408,60 +450,114 @@ int netif_type(int sockfd, struct ifaddrs *ifaddr, struct ifreq *ifr) {
 
 
 // handle aggregated interfaces
-void netif_bond(int sockfd, struct netif *netifs, struct netif *master) {
+void netif_bond(int sockfd, struct netif *netifs, struct netif *master,
+		struct ifreq *ifr) {
 
     struct netif *subif = NULL, *csubif = master;
     int i;
 
-#ifdef HAVE_LINUX_IF_BONDING_H
+#ifdef HAVE_SYSFS
     // handle linux bonding interfaces
     char path[SYSFS_PATH_MAX];
-    FILE *fp;
+    FILE *file;
     char line[1024];
     char *slave, *nslave;
+#endif /* HAVE_SYSFS */
+
+#ifdef HAVE_LINUX_IF_BONDING_H
+    struct ifbond ifbond;
+    struct ifslave ifslave;
+#endif /* HAVE_LINUX_IF_BONDING_H */
 
     // check for lacp
-    sprintf(path, "%s/%s/bonding/mode", SYSFS_VIRTUAL, master->name); 
-    if ((fp = fopen(path, "r")) != NULL) {
-	if (fscanf(fp, "802.3ad") != EOF)
-	    master->lacp = 1;
-	fclose(fp);
+
+#ifdef HAVE_SYSFS
+    // via sysfs
+    if (snprintf(path, SYSFS_PATH_MAX, "%s/%s/bonding/mode", 
+		SYSFS_CLASS_NET, master->name) > 0) {
+	if ((file = fopen(path, "r")) != NULL) {
+	    if (fscanf(file, "802.3ad") != EOF)
+		master->lacp = 1;
+	    (void) fclose(file);
+	}
     }
+#endif /* HAVE_SYSFS */
+
+#ifdef HAVE_LINUX_IF_BONDING_H
+    strncpy(ifr->ifr_name, master->name, IFNAMSIZ);
+    memset(&ifbond, 0, sizeof(ifbond));
+    ifr->ifr_data = (char *)&ifbond;
+
+    if (ioctl(sockfd, SIOCBONDINFOQUERY, ifr) >= 0) {
+	if (ifbond.bond_mode == BOND_MODE_8023AD)
+	    master->lacp = 1;
+    }
+#endif /* HAVE_LINUX_IF_BONDING_H */
+
+    if (master->lacp == 1)
+	my_log(INFO, "lacp enabled on %s", master->name);
+
 
     // handle slaves
-    sprintf(path, "%s/%s/bonding/slaves", SYSFS_VIRTUAL, master->name); 
-    if ((fp = fopen(path, "r")) != NULL) {
-	if (fgets(line, sizeof(line), fp) != NULL) {
-	    // remove newline
-	    *strchr(line, '\n') = '\0';
 
-	    slave = line;
-	    i = 0;
-	    while (strlen(slave) > 0) {
-		nslave = strstr(line, " ");
-		if (nslave != NULL)
-		    *nslave = '\0';
+#ifdef HAVE_SYSFS
+    // via sysfs
+    if ((snprintf(path, SYSFS_PATH_MAX,
+	    SYSFS_CLASS_NET "/%s/bonding/slaves", master->name) > 0) &&
+	(read_line(path, line, sizeof(line)) != -1)) {
 
-		subif = netif_byname(netifs, slave);
-		if (subif != NULL) {
-		    my_log(3, "found slave %s", subif->name);
-		    subif->slave = 1;
-		    subif->master = master;
-		    subif->lacp_index = i++;
-		    csubif->subif = subif;
-		    csubif = subif;
-		}
+	slave = line;
+	i = 0;
+	while (strlen(slave) > 0) {
+	    nslave = strstr(slave, " ");
+	    if (nslave != NULL)
+		*nslave = '\0';
 
-		if (nslave != NULL) {
-		    nslave++;
-		    slave = nslave;
-		} else {
-		    break;
-		}
+	    subif = netif_byname(netifs, slave);
+	    if (subif != NULL) {
+		my_log(INFO, "found slave %s", subif->name);
+		subif->slave = 1;
+		subif->master = master;
+		subif->lacp_index = i++;
+		csubif->subif = subif;
+		csubif = subif;
 	    }
-	};
 
-	fclose(fp);
+	    if (nslave != NULL) {
+		nslave++;
+		slave = nslave;
+	    } else {
+		break;
+	    }
+	}
+
+	return;
+    }
+#endif /* HAVE_SYSFS */
+
+#ifdef HAVE_LINUX_IF_BONDING_H
+    // or ioctl
+
+    // check for a sensible num_slaves entry
+    if (ifbond.num_slaves <= 0)
+	return;
+
+    for (i = 0; i < ifbond.num_slaves; i++) {
+	ifslave.slave_id = i;
+	ifr->ifr_data = (char *)&ifslave;
+
+	if (ioctl(sockfd, SIOCBONDSLAVEINFOQUERY, ifr) >= 0) {
+	    subif = netif_byname(netifs, ifslave.slave_name);
+
+	    if (subif != NULL) {
+		my_log(INFO, "found slave %s", subif->name);
+		subif->slave = 1;
+		subif->master = master;
+		subif->lacp_index = i;
+		csubif->subif = subif;
+		csubif = subif;
+	    }
+	}
     }
 
     return;
@@ -478,7 +574,7 @@ void netif_bond(int sockfd, struct netif *netifs, struct netif *master) {
 
     memset(&ra, 0, sizeof(ra));
 
-    strncpy(ra.ra_ifname, master->name, sizeof(ra.ra_ifname));
+    strlcpy(ra.ra_ifname, master->name, sizeof(ra.ra_ifname));
     ra.ra_size = sizeof(rpbuf);
     ra.ra_port = rpbuf;
 
@@ -497,7 +593,7 @@ void netif_bond(int sockfd, struct netif *netifs, struct netif *master) {
 	subif = netif_byname(netifs, rpbuf[i].rp_portname);
 
 	if (subif != NULL) {
-	    my_log(3, "found slave %s", subif->name);
+	    my_log(INFO, "found slave %s", subif->name);
 	    subif->slave = 1;
 	    subif->master = master;
 	    subif->lacp_index = i++;
@@ -512,28 +608,67 @@ void netif_bond(int sockfd, struct netif *netifs, struct netif *master) {
 
 
 // handle bridge interfaces
-void netif_bridge(int sockfd, struct netif *netifs, struct netif *master) {
+void netif_bridge(int sockfd, struct netif *netifs, struct netif *master,
+		  struct ifreq *ifr) {
 
     struct netif *subif = NULL, *csubif = master;
+    int i;
 
-#if HAVE_LINUX_IF_BRIDGE_H 
+#if HAVE_SYSFS
     // handle linux bridge interfaces
     char path[SYSFS_PATH_MAX];
     DIR  *dir;
     struct dirent *dirent;
+#endif /* HAVE_SYSFS */
+
+#ifdef HAVE_LINUX_IF_BRIDGE_H
+    int ifindex[BRIDGE_MAX_PORTS];
+    unsigned long args[4] = { BRCTL_GET_PORT_LIST,
+		    (unsigned long)ifindex, BRIDGE_MAX_PORTS, 0 };
+#endif /* HAVE_LINUX_IF_BRIDGE_H */
+
 
     // handle slaves
-    sprintf(path, SYSFS_VIRTUAL "/%s/" SYSFS_BRIDGE_PORT_SUBDIR, master->name); 
+ 
+#if HAVE_SYSFS
+    // via sysfs
+    sprintf(path, SYSFS_CLASS_NET "/%s/" SYSFS_BRIDGE_PORT_SUBDIR, master->name); 
+    if ((dir = opendir(path)) != NULL) {
 
-    if ((dir = opendir(path)) == NULL) {
-	my_log(0, "reading bridge %s subdir %s failed: %s",
-	    master->name, path, strerror(errno));
+	while ((dirent = readdir(dir)) != NULL) {
+	    subif = netif_byname(netifs, dirent->d_name);
+
+	    if (subif != NULL) {
+		my_log(INFO, "found slave %s", subif->name);
+		subif->slave = 1;
+		subif->master = master;
+		csubif->subif = subif;
+		csubif = subif;
+	    }
+	}
+
+	closedir(dir);
+	return;
+    }
+#endif /* HAVE_SYSFS */
+
+#ifdef HAVE_LINUX_IF_BRIDGE_H
+    // or ioctl
+    memset(ifindex, 0, sizeof(ifindex));
+    strncpy(ifr->ifr_name, master->name, IFNAMSIZ);
+    ifr->ifr_data = (char *)&args;
+
+    if (ioctl(sockfd, SIOCDEVPRIVATE, ifr) < 0) {
+	my_log(CRIT, "bridge ioctl failed on interface %s: %s",
+	       master->name, strerror(errno));
 	return;
     }
 
-    while ((dirent = readdir(dir)) != NULL) {
-	subif = netif_byname(netifs, dirent->d_name);
+    for (i = 0; i < BRIDGE_MAX_PORTS; i++) {
+	subif = netif_byindex(netifs, ifindex[i]);
+
 	if (subif != NULL) {
+	    my_log(INFO, "found slave %s", subif->name);
 	    subif->slave = 1;
 	    subif->master = master;
 	    csubif->subif = subif;
@@ -541,7 +676,6 @@ void netif_bridge(int sockfd, struct netif *netifs, struct netif *master) {
 	}
     }
 
-    closedir(dir);
     return;
 #endif /* HAVE_LINUX_IF_BRIDGE_H */
 
@@ -549,14 +683,14 @@ void netif_bridge(int sockfd, struct netif *netifs, struct netif *master) {
     struct ifbifconf bifc;
     struct ifbreq *req;
     char *inbuf = NULL, *ninbuf;
-    int i, len = 8192;
+    int len = 8192;
 
 #ifdef HAVE_NET_IF_BRIDGEVAR_H
     struct ifdrv ifd;
 
     memset(&ifd, 0, sizeof(ifd));
 
-    strncpy(ifd.ifd_name, master->name, sizeof(ifd.ifd_name));
+    strlcpy(ifd.ifd_name, master->name, sizeof(ifd.ifd_name));
     ifd.ifd_cmd = BRDGGIFS;
     ifd.ifd_len = sizeof(bifc);
     ifd.ifd_data = &bifc;
@@ -568,7 +702,7 @@ void netif_bridge(int sockfd, struct netif *netifs, struct netif *master) {
 	if (ninbuf == NULL) {
 	    if (inbuf != NULL)
 		free(inbuf);
-	    my_log(1, "unable to allocate interface buffer");
+	    my_log(WARN, "unable to allocate interface buffer");
 	    return;
 	}
 
@@ -594,6 +728,7 @@ void netif_bridge(int sockfd, struct netif *netifs, struct netif *master) {
 
 	subif = netif_byname(netifs, req->ifbr_ifsname);
 	if (subif != NULL) {
+	    my_log(INFO, "found slave %s", subif->name);
 	    subif->slave = 1;
 	    subif->master = master;
 	    csubif->subif = subif;
@@ -611,7 +746,8 @@ void netif_bridge(int sockfd, struct netif *netifs, struct netif *master) {
 
 
 // perform address detection for all netifs
-int netif_addrs(struct ifaddrs *ifaddrs, struct netif *netifs) {
+int netif_addrs(struct ifaddrs *ifaddrs, struct netif *netifs,
+		struct sysinfo *sysinfo) {
     struct ifaddrs *ifaddr;
     struct netif *netif;
 
@@ -625,6 +761,11 @@ int netif_addrs(struct ifaddrs *ifaddrs, struct netif *netifs) {
 #endif
 
     for (ifaddr = ifaddrs; ifaddr != NULL; ifaddr = ifaddr->ifa_next) {
+
+	// skip interfaces without addresses
+	if (ifaddr->ifa_addr == NULL)
+	    continue;
+
 	// fetch the netif for this ifaddr
 	netif = netif_byname(netifs, ifaddr->ifa_name);
 	if (netif == NULL)
@@ -672,15 +813,32 @@ int netif_addrs(struct ifaddrs *ifaddrs, struct netif *netifs) {
 	}
     }
 
+    // return when no management addresses are defined
+    if ((sysinfo->maddr4 == 0) &&
+	(IN6_IS_ADDR_UNSPECIFIED((struct in6_addr *)sysinfo->maddr6)) )
+	return(EXIT_SUCCESS);
+
+    // use management address when unnumbered
+    for (netif = netifs; netif != NULL; netif = netif->next) {
+
+	if ((netif->ipaddr4 == 0) || (sysinfo->maddr_force == 1))
+	    netif->ipaddr4 = sysinfo->maddr4;
+
+	if (IN6_IS_ADDR_UNSPECIFIED((struct in6_addr *)netif->ipaddr6) ||
+	    (sysinfo->maddr_force == 1))
+	    memcpy(&netif->ipaddr6, &sysinfo->maddr6, sizeof(sysinfo->maddr6));
+    }
+
     return(EXIT_SUCCESS);
 }
 
 
 // detect forwarding capability
-int netif_forwarding() {
+void netif_forwarding(struct sysinfo *sysinfo) {
 
-    FILE *file;
+#ifdef HAVE_PROC_SYS_NET
     char line[256];
+#endif
 
 #ifdef CTL_NET
     int mib[4], n;
@@ -691,37 +849,51 @@ int netif_forwarding() {
     mib[0] = CTL_NET;
 #endif
 
-    if ((file = fopen(PROCFS_FORWARD_IPV4, "r")) != NULL) {
-        if (fgets(line, sizeof(line), file))
-            if (atoi(line) == 1)
-		return(1);
-	fclose(file);
+#ifdef HAVE_PROC_SYS_NET
+    if (read_line(PROCFS_FORWARD_IPV4, line, sizeof(line)) != -1) {
+	sysinfo->cap |= CAP_ROUTER; 
+
+        if (atoi(line) == 1) {
+	    sysinfo->cap_active |= CAP_ROUTER; 
+	    return;
+	}
     }
 
-    if ((file = fopen(PROCFS_FORWARD_IPV6, "r")) != NULL) {
-        if (fgets(line, sizeof(line), file))
-            if (atoi(line) == 1)
-		return(1);
-	fclose(file);
+    if (read_line(PROCFS_FORWARD_IPV6, line, sizeof(line)) != -1) {
+	sysinfo->cap |= CAP_ROUTER; 
+
+        if (atoi(line) == 1) {
+	    sysinfo->cap_active |= CAP_ROUTER; 
+	    return;
+	}
     }
+#endif
 
 #ifdef CTL_NET
     mib[1] = PF_INET;
     mib[2] = IPPROTO_IP;
     mib[3] = IPCTL_FORWARDING;
 
-    if (sysctl(mib, 4, &n, &len, NULL, 0) != -1 && n)
-	return(1);
+    if (sysctl(mib, 4, &n, &len, NULL, 0) != -1) {
+	sysinfo->cap |= CAP_ROUTER; 
+	if (n == 1) {
+	    sysinfo->cap_active |= CAP_ROUTER; 
+	    return;
+	}
+    }
 
     mib[1] = PF_INET6;
     mib[2] = IPPROTO_IPV6;
     mib[3] = IPV6CTL_FORWARDING;
 
-    if (sysctl(mib, 4, &n, &len, NULL, 0) != -1 && n)
-	return(1);
+    if (sysctl(mib, 4, &n, &len, NULL, 0) != -1) {
+	sysinfo->cap |= CAP_ROUTER; 
+	if (n == 1) {
+	    sysinfo->cap_active |= CAP_ROUTER; 
+	    return;
+	}
+    }
 #endif
-
-    return(0);
 }
 
 
@@ -747,13 +919,13 @@ int netif_media(struct netif *netif) {
     netif->mau = 0;
 
     memset(&ifr, 0, sizeof(ifr));
-    strncpy(ifr.ifr_name, netif->name, IFNAMSIZ);
+    strlcpy(ifr.ifr_name, netif->name, sizeof(ifr.ifr_name));
 
     // interface mtu
     if (ioctl(sockfd, SIOCGIFMTU, (caddr_t)&ifr) >= 0)
 	netif->mtu = ifr.ifr_mtu;
     else
-	my_log(3, "mtu detection failed on interface %s", netif->name);
+	my_log(INFO, "mtu detection failed on interface %s", netif->name);
 
 #if HAVE_LINUX_ETHTOOL_H
     memset(&ecmd, 0, sizeof(ecmd));
@@ -766,29 +938,43 @@ int netif_media(struct netif *netif) {
 
 	// autoneg
 	if (ecmd.supported & SUPPORTED_Autoneg) {
-	    my_log(3, "autoneg supported on %s", netif->name);
+	    my_log(INFO, "autoneg supported on %s", netif->name);
 	    netif->autoneg_supported = 1;
 	    netif->autoneg_enabled = (ecmd.autoneg == AUTONEG_ENABLE) ? 1 : 0;
 	} else {
-	    my_log(3, "autoneg not supported on %s", netif->name);
+	    my_log(INFO, "autoneg not supported on %s", netif->name);
 	    netif->autoneg_supported = 0;
 	}	
     } else {
-	my_log(3, "ethtool ioctl failed on interface %s", netif->name);
+	my_log(INFO, "ethtool ioctl failed on interface %s", netif->name);
     }
 #endif /* HAVE_LINUX_ETHTOOL_H */
 
 #if HAVE_NET_IF_MEDIA_H
     memset(&ifmr, 0, sizeof(ifmr));
-    strncpy(ifmr.ifm_name, netif->name, IFNAMSIZ);
+    strlcpy(ifmr.ifm_name, netif->name, sizeof(ifmr.ifm_name));
 
     if (ioctl(sockfd, SIOCGIFMEDIA, (caddr_t)&ifmr) < 0) {
-	my_log(3, "media detection not supported on %s", netif->name);
+	my_log(INFO, "media detection not supported on %s", netif->name);
+	close(sockfd);
+	return(EXIT_SUCCESS);
+    }
+
+    if (IFM_TYPE(ifmr.ifm_current) != IFM_ETHER) {
+	my_log(INFO, "non-ethernet interface %s found", netif->name);
+	close(sockfd);
+	return(EXIT_FAILURE);
+    }
+
+    if ((ifmr.ifm_status & IFM_ACTIVE) == 0) { 
+	my_log(INFO, "no link detected on interface %s", netif->name);
+	close(sockfd);
 	return(EXIT_SUCCESS);
     }
 
     if (ifmr.ifm_count == 0) {
-	my_log(0, "missing media types for interface %s", netif->name);
+	my_log(CRIT, "missing media types for interface %s", netif->name);
+	close(sockfd);
 	return(EXIT_FAILURE);
     }
 
@@ -796,24 +982,16 @@ int netif_media(struct netif *netif) {
     ifmr.ifm_ulist = media_list;
 
     if (ioctl(sockfd, SIOCGIFMEDIA, (caddr_t)&ifmr) < 0) {
-	my_log(0, "media detection failed for interface %s", netif->name);
+	my_log(CRIT, "media detection failed for interface %s", netif->name);
+	free(media_list);
+	close(sockfd);
 	return(EXIT_FAILURE);
     }
 
-    if (IFM_TYPE(ifmr.ifm_current) != IFM_ETHER) {
-	my_log(0, "non-ethernet interface %s found", netif->name);
-	return(EXIT_FAILURE);
-    }
-
-    if ((ifmr.ifm_status & IFM_ACTIVE) == 0) { 
-	my_log(0, "no link detected on interface %s", netif->name);
-	return(EXIT_SUCCESS);
-    }
-
     // autoneg support
     for (i = 0; i < ifmr.ifm_count; i++) {
 	if (IFM_SUBTYPE(ifmr.ifm_ulist[i]) == IFM_AUTO) {
-	    my_log(3, "autoneg supported on %s", netif->name);
+	    my_log(INFO, "autoneg supported on %s", netif->name);
 	    netif->autoneg_supported = 1;
 	    break;
 	}
@@ -822,23 +1000,23 @@ int netif_media(struct netif *netif) {
     // autoneg enabled
     if (netif->autoneg_supported == 1) {
 	if (IFM_SUBTYPE(ifmr.ifm_current) == IFM_AUTO) {
-	    my_log(3, "autoneg enabled on %s", netif->name);
+	    my_log(INFO, "autoneg enabled on %s", netif->name);
 	    netif->autoneg_enabled = 1;
 	} else {
-	    my_log(3, "autoneg disabled on interface %s", netif->name);
+	    my_log(INFO, "autoneg disabled on interface %s", netif->name);
 	    netif->autoneg_enabled = 0;
 	}
     } else {
-	my_log(3, "autoneg not supported on interface %s", netif->name);
+	my_log(INFO, "autoneg not supported on interface %s", netif->name);
 	netif->autoneg_supported = 0;
     }
 
     // duplex
     if ((IFM_OPTIONS(ifmr.ifm_active) & IFM_FDX) != 0) {
-	my_log(3, "full-duplex enabled on interface %s", netif->name);
+	my_log(INFO, "full-duplex enabled on interface %s", netif->name);
 	netif->duplex = 1;
     } else {
-	my_log(3, "half-duplex enabled on interface %s", netif->name);
+	my_log(INFO, "half-duplex enabled on interface %s", netif->name);
 	netif->duplex = 0;
     }
 
